import json
from dataclasses import dataclass
from datetime import datetime
from json import JSONDecodeError
from pathlib import Path
from typing import Any, List, Set

import chess
import pandas as pd
from omegaconf import DictConfig
from rich.console import Console
from rich.table import Table

from chessli import ChessliPaths, users_client, utils
from chessli.enums import PuzzleDBSource
from chessli.rich_logging import log
from chessli.utils import import_to_anki_via_apy, in_bold

console = Console()


@dataclass
class PuzzleFetcherMixin:
    config: DictConfig
    paths: ChessliPaths

    def fetch_puzzle_activity(self):
        log.info(f"Fetching new puzzle activity...")
        puzzle_activity = list(users_client.get_puzzle_activity())
        return puzzle_activity

    def read_lichess_puzzle_database(self) -> pd.DataFrame:
        column_names = "PuzzleId,FEN,Moves,Rating,RatingDeviation,Popularity,NbPlays,Themes,GameUrl".split(
            ","
        )
        if self.config.db_source == PuzzleDBSource.remote:
            url = "https://database.lichess.org/lichess_db_puzzle.csv.bz2"
            log.info(
                f"Trying to read the most up-to-date lichess puzzle database from {url}. This may take a few seconds..."
            )
            puzzle_df = pd.read_csv(url, names=column_names, compression="bz2")
        elif self.config.db_source == PuzzleDBSource.local:
            puzzle_db_path = self.paths.tactics_dir / "lichess_db_puzzle.csv"
            log.info(
                f"Trying to read the lichess puzzle database from {puzzle_db_path}."
            )
            puzzle_df = pd.read_csv(puzzle_db_path, names=column_names)
        else:
            raise NotImplementedError(
                f"Unknown puzzle database source {self.config.db_source}"
            )
        return puzzle_df


@dataclass
class TacticsManager(PuzzleFetcherMixin, object):
    config: DictConfig
    paths: ChessliPaths

    def get_puzzle_ids_path(self) -> Path:
        puzzles_ids_path = self.paths.tactics_dir / "played_puzzles_ids.json"
        puzzles_ids_path.touch(exist_ok=True)
        return puzzles_ids_path

    def read_puzzle_ids(self) -> List:
        puzzle_ids_path = self.get_puzzle_ids_path()
        try:
            with puzzle_ids_path.open("r") as fp:
                old_puzzle_ids = json.load(fp)
        except (FileNotFoundError, JSONDecodeError):
            old_puzzle_ids = []
        return old_puzzle_ids

    def store_puzzle_ids(self, puzzle_ids: List[str]) -> None:
        puzzle_ids_path = self.get_puzzle_ids_path()
        with puzzle_ids_path.open("w") as fp:
            json.dump(puzzle_ids, fp)

    def _get_ids_from_puzzle_activity(
        self, puzzle_activity: List[Any], new_only: bool = True,
    ) -> List[str]:
        puzzle_ids = [puzzle["id"] for puzzle in puzzle_activity]
        if new_only:
            old_puzzle_ids = self.read_puzzle_ids()
            new_puzzle_ids = set(puzzle_ids) - set(old_puzzle_ids)
            log.info(f"There are {len(new_puzzle_ids)} new puzzles!")
            return new_puzzle_ids
        else:
            return puzzle_ids

    def print_new_puzzles(self) -> None:
        puzzle_activity = self.fetch_puzzle_activity()
        puzzle_ids = self._get_ids_from_puzzle_activity(puzzle_activity)
        if not puzzle_ids:
            return None
        puzzles_df = self.read_lichess_puzzle_database()
        new_puzzles_df = self._extract_new_puzzles(puzzle_ids, puzzles_df)

        table = Table(
            *list(new_puzzles_df),
            title=f"Played Puzzles ({len(new_puzzles_df)}) :fire:",
        )
        for idx, puzzle in new_puzzles_df.iterrows():
            table.add_row(*[str(val) for val in puzzle.values])

        console.print(table)

    def _extract_new_puzzles(self, puzzle_ids, df: pd.DataFrame) -> pd.DataFrame:
        new_puzzles = df.loc[df["PuzzleId"].isin(puzzle_ids)]

        def assign_san_moves(df: pd.DataFrame) -> List[str]:
            """We need to do some transformation to get a compatible move list"""
            move_list = []
            for idx, row in df.iterrows():
                san_move = chess.Board(row["FEN"]).variation_san(
                    [chess.Move.from_uci(m) for m in row["Moves"].split()]
                )
                move_list.append(san_move)

            return move_list

        pd.options.mode.chained_assignment = None  # Suppress false positive warning
        new_puzzles["Move List"] = assign_san_moves(new_puzzles)
        return new_puzzles

    def update_stored_puzzle_ids(self, puzzle_ids) -> None:
        old_puzzle_ids = self.read_puzzle_ids()
        if not old_puzzle_ids:
            old_puzzle_ids = []
        puzzle_ids = set(puzzle_ids) | set(old_puzzle_ids)
        self.store_puzzle_ids(list(puzzle_ids))

    def ankify_puzzles(self) -> None:
        if self.config.fetch:
            puzzle_activity = self.fetch_puzzle_activity()
            puzzle_ids = self._get_ids_from_puzzle_activity(puzzle_activity)
        else:
            puzzle_ids = self.read_puzzle_ids()

        if not puzzle_ids:
            log.info(
                f"There are no new puzzles to be ankified! Time to play some! :fire:"
            )
            return

        puzzles_df = self.read_lichess_puzzle_database()
        new_puzzles_df = self._extract_new_puzzles(puzzle_ids, puzzles_df)

        time_stamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

        apy_str = self.puzzle_df_to_apy_md(new_puzzles_df)
        last_puzzles_path = self.paths.tactics_dir / f"puzzle_export_{time_stamp}.md"

        last_puzzles_path.write_text(apy_str)

        new_puzzles_df.to_csv(
            index=False, path_or_buf=last_puzzles_path.with_suffix(".csv")
        )

        log.info(
            f"Exported played puzzles to {in_bold(last_puzzles_path)} and {in_bold(last_puzzles_path.with_suffix('.csv'))}"
        )

        if not self.config.export_only:
            log.info(f"Firing up 'apy' to import the new puzzles into anki.")
            import_to_anki_via_apy(file_path=last_puzzles_path)

        self.update_stored_puzzle_ids(puzzle_ids)

    @staticmethod
    def puzzle_df_to_apy_md(df: pd.DataFrame) -> str:
        md = ""
        header = "model: Chessli Tactics\ntags: chess::tactics\ndeck: Chessli::tactics\nmarkdown: False\n\n"
        md += header
        for idx, row in df.iterrows():
            md += "# Note\n"
            for key, value in row.items():
                md += f"## {key}\n"
                md += f"{value}\n"
        return md



from datetime import datetime

import typer

from chessli import __version__
from chessli.enums import SinceEnum
from chessli.utils import as_title, convert_since_enum_to_millis


def version_callback(value: bool):
    if value:
        print(as_title(f"Chessli {__version__}"))
        raise typer.Exit()


def since_callback(ctx: typer.Context, value: SinceEnum):
    if value == SinceEnum.last_time:
        chessli_paths = ctx.parent.params["paths"]
        if chessli_paths.user_config.last_fetch_time is None:
            chessli_paths.user_config.last_fetch_time = str(datetime.now())
            ctx.parent.params["paths"] = chessli_paths
    return value.value



from enum import Enum

__all__ = ["PerfType", "Variant", "Color", "Room", "Mode", "Position"]


class StrEnum(str, Enum):
    """A String enum"""


class PuzzleDBSource(str, Enum):
    remote = "remote"
    local = "local"


class SinceEnum(StrEnum):
    last_time = "last-time"
    one_hour = "one-hour"
    yesterday = "yesterday"
    last_week = "last-week"
    last_month = "last-month"
    last_year = "last-year"
    forever = "forever"


class Nag(Enum):
    good = 1  # A good move.  by ``!`` in PGN notation.
    mistake = 2  # A mistake.  by ``?`` in PGN notation.
    brilliant = 3  # A brilliant move.  by ``!!`` in PGN notation.
    blunder = 4  # A blunder.  by ``??`` in PGN notation.
    speculative = 5  # A speculative move. ``!?`` in PGN notation.
    dubious = 6  # A dubious move.  by ``?!`` in PGN notation.


class GameType(StrEnum):
    ANTICHESS = "antichess"
    ATOMIC = "atomic"
    CHESS960 = "chess960"
    CRAZYHOUSE = "crazyhouse"
    HORDE = "horde"
    KING_OF_THE_HILL = "kingOfTheHill"
    RACING_KINGS = "racingKings"
    THREE_CHECK = "threeCheck"


class PerfType(StrEnum):
    ANTICHESS = GameType.ANTICHESS.value
    ATOMIC = GameType.ATOMIC.value
    CHESS960 = GameType.CHESS960.value
    CRAZYHOUSE = GameType.CRAZYHOUSE.value
    HORDE = GameType.HORDE.value
    KING_OF_THE_HILL = GameType.KING_OF_THE_HILL.value
    RACING_KINGS = GameType.RACING_KINGS.value
    THREE_CHECK = GameType.THREE_CHECK.value

    BULLET = "bullet"
    BLITZ = "blitz"
    RAPID = "rapid"
    CLASSICAL = "classical"
    ULTRA_BULLET = "ultraBullet"


class Color(Enum):
    white = 1
    black = 0



import typer
from rich import print

from chessli.enums import PuzzleDBSource
from chessli.tactics import TacticsManager
from chessli.utils import (
    as_title,
    create_config_from_options,
    extract_context_info,
    in_bold,
)

app = typer.Typer()


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    db_source: PuzzleDBSource = typer.Option(
        PuzzleDBSource.remote,
        help="Select where to get the lichess puzzle database from.",
    ),
):
    """Chessli Tactics & Puzzles"""
    ctx.params = {**ctx.params, **ctx.parent.params}
    print(f"{as_title('chessli tactics')}", end="\n\n")


@app.command()
def ls(
    ctx: typer.Context,
    new: bool = typer.Option(
        True, "--new/--old", help="Select whether to fetch and list new puzzles only"
    ),
):
    """Print a pretty table of the newly played puzzles"""
    chessli_paths, config = extract_context_info(ctx)
    tactics_manager = TacticsManager(config, chessli_paths)
    if new:
        tactics_manager.print_new_puzzles()
    else:
        puzzle_ids = tactics_manager.read_puzzle_ids()
        print(f"{puzzle_ids}")


@app.command()
def ankify(
    ctx: typer.Context,
    new: bool = typer.Option(
        True,
        "--new/--all",
        help="Select whether to only ankify new puzzles or all puzzles",
    ),
    export_only: bool = typer.Option(
        True,
        "--export-only/--directly",
        help="Select to only export the created anki cards",
    ),
):
    """Optionally fetch new puzzles and ankify them"""
    chessli_paths, config = extract_context_info(ctx)

    tactics_manager = TacticsManager(config, chessli_paths)
    tactics_manager.ankify_puzzles()


if __name__ == "__main__":
    app()



from typing import Callable, Dict, List, Optional, Set, Tuple, Union

import typer
from omegaconf import OmegaConf
from rich import print
from rich.console import Console
from rich.progress import track

from chessli.cli.option_callbacks import since_callback
from chessli.enums import PerfType, SinceEnum
from chessli.games import GamesCollection, GamesFetcher, GamesReader
from chessli.utils import (
    as_title,
    convert_since_enum_to_millis,
    create_config_from_options,
    extract_context_info,
)

app = typer.Typer()
console = Console()


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context,):
    """Fetch, store, show and ankify games and mistakes"""
    ctx.params = ctx.parent.params
    print(as_title("chessli games"), end="\n\n")


@app.command()
def ls(
    ctx: typer.Context,
    perf_type: Optional[PerfType] = typer.Option(
        None, help="Filter games to the selected `perf-type`"
    ),
):
    """List your games"""
    chessli_paths, cli_config = extract_context_info(ctx)

    games_reader = GamesReader(chessli_paths, cli_config)
    games_reader.ls()


@app.command()
def fetch(
    ctx: typer.Context,
    verbose: int = typer.Option(
        1, "--verbose", "-v", count=True, help="Select the verbosity level"
    ),
    perf_type: Optional[List[PerfType]] = typer.Option(
        [], help="Filter fetching of games to the selected `perf_types`"
    ),
    since_enum: SinceEnum = typer.Option(
        SinceEnum.last_time,
        "--since",
        help="Filter fetching of games to those played since `since`",
        callback=since_callback,
    ),
    max: Optional[int] = typer.Option(30, help="Limit fetching of games to `max`",),
    store: bool = typer.Option(False, help="Select if fetched games should be stored"),
):
    """Fetch games from lichess and optionally store them"""
    chessli_paths, cli_config = extract_context_info(ctx)

    cli_config["since_millis"] = convert_since_enum_to_millis(
        since_enum, chessli_paths.user_config
    )

    games_fetcher = GamesFetcher(chessli_paths, cli_config)
    new_games = games_fetcher.fetch_games()


@app.command()
def ankify(
    ctx: typer.Context,
    new_games_only: bool = typer.Option(
        True, help="Fetch new games and only ankify those"
    ),
    since_enum: SinceEnum = typer.Option(
        SinceEnum.last_time,
        "--since",
        help="Filter fetching of games to those played since `since`",
        callback=since_callback,
    ),
    max: Optional[int] = typer.Option(30, help="Limit fetching of games to `max`",),
    perf_type: Optional[List[PerfType]] = typer.Option(
        None, help="Filter fetching of games to the selected `perf_types`"
    ),
    export_only: bool = typer.Option(
        True,
        "--export-only/--directly",
        help="Select to only export the created anki cards",
    ),
):
    """Parse your games to find mistakes and create Anki cards"""

    chessli_paths, cli_config = extract_context_info(ctx)
    cli_config["since_millis"] = convert_since_enum_to_millis(
        since_enum, chessli_paths.user_config
    )

    if new_games_only:
        games = GamesFetcher(chessli_paths, cli_config).fetch_games()
    else:
        games = GamesReader(chessli_paths, cli_config).games

    games_collection = GamesCollection(
        config=cli_config, paths=chessli_paths, games=games
    )
    if export_only:
        games_collection.export_csv()
    else:
        games_collection.ankify_games()


if __name__ == "__main__":
    app()



from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import chess
from omegaconf import DictConfig
from rich import print
from rich.console import Console
from rich.markdown import Markdown

from chessli.enums import Color, Nag

console = Console()


def get_nag_name(nag: int) -> str:
    return {nag.value: nag.name for nag in Nag}[nag]


@dataclass
class Mistake:
    fen: str  # The fen of the board serves as ID
    comment: str
    ply: int
    move_color: str
    nag: Nag
    nag_name: str
    my_move: str
    best_move: str
    variation: str
    game: Any

    @property
    def game_items(self) -> Dict:
        return {k: v for k, v in self.game.headers.items()}

    @property
    def items(self) -> Dict:
        raw_items = {**vars(self), **self.game_items}
        dont_include = [
            "WhiteElo",
            "WhiteTitle",
            "WhiteRatingDiff",
            "UTCDate",
            "game",
            "BlackTitle",
            "BlackRatingDiff",
            "BlackElo",
            "UTCTime",
        ]
        items = {k: v for k, v in raw_items.items() if k not in dont_include}
        if items["move_color"] == "White":
            items["Me"] = raw_items["White"]
            items["Opponent"] = raw_items["Black"]
            items["My Elo"] = raw_items["WhiteElo"]
            items["Opponent Elo"] = raw_items["BlackElo"]
        elif items["move_color"] == "Black":
            items["Me"] = raw_items["Black"]
            items["Opponent"] = raw_items["White"]
            items["My Elo"] = raw_items["BlackElo"]
            items["Opponent Elo"] = raw_items["WhiteElo"]

        return items

    @property
    def md(self) -> str:
        md = "# Mistake\n"
        for key, value in self.items.items():
            if key == "game":
                continue
            md += f"## {key.replace('_', ' ').title()}\n"
            md += f"{value}\n"
        return md

    @property
    def pprint(self):
        console.print(Markdown(self.md))


@dataclass
class MistakeFinderMixin:
    pgn: Optional[chess.pgn.Game] = None
    config: Optional[DictConfig] = None

    @property
    def player(self,) -> Color:
        return (
            Color.white
            if self.pgn.headers["White"] == self.config.user
            else Color.black
        )

    def my_move(self, move_color: Color) -> bool:
        return True if self.player.value == move_color else False

    @property
    def mistakes(self,) -> List["Mistake"]:
        _mistakes = []

        game = self.pgn

        while game is not None:
            whose_turn = not game.turn()
            move_color = "White" if whose_turn else "Black"
            if self.my_move(whose_turn):
                if game.nags:
                    parent = game.parent
                    parent_board = parent.board()
                    if len(parent.variations) > 1:
                        assert len(parent.variations) == 2
                        variation_moves = str(parent.variations[1])
                    nag = list(game.nags)[0]
                    nag_name = get_nag_name(nag)

                    _mistakes.append(
                        Mistake(
                            fen=parent_board.fen(),
                            comment=game.comment,
                            ply=parent.ply(),
                            move_color=move_color,
                            nag=nag,
                            nag_name=nag_name,
                            my_move=game.san(),
                            best_move=parent.variations[1].san(),
                            variation=variation_moves,
                            game=self.pgn,
                        )
                    )
            game = game.next()

        return _mistakes



from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, ItemsView, List, Optional, Set, Tuple, Union

import chess
import pandas as pd
from omegaconf import DictConfig
from rich import print
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table

from chessli import ChessliPaths
from chessli.rich_logging import log
from chessli.utils import import_to_anki_via_apy, in_bold

console = Console()


class ECOVolume(str, Enum):
    A = "Volume A: Flank openings"
    B = "Volume B: Semi-Open Games other than the French Defense"
    C = "Volume C: Open Games and the French Defense"
    D = "Volume D: Closed Games and Semi-Closed Games"
    E = "Volume E: Indian Defenses"


class ECOVolumeLetter(str, Enum):
    A = "A"
    B = "B"
    C = "C"
    D = "D"
    E = "E"


@dataclass
class Opening:
    name: str
    site: str
    eco: str
    moves: str
    config: Any
    paths: ChessliPaths

    def __str__(self):
        return f"{self.eco} - {self.name}"

    @property
    def items(self) -> Dict:
        return vars(self)

    @property
    def anki_items(self) -> Dict:
        anki_items = {
            k: v for k, v in vars(self).items() if k not in ["config", "paths"]
        }
        return anki_items

    @property
    def md(self) -> str:
        md = "# Opening\n"
        for key, value in self.anki_items.items():
            md += f"## {key.replace('_', ' ').title()}\n"
            md += f"{value}\n"
        return md

    @property
    def pprint(self) -> None:
        console.print(Markdown(self.md))

    @property
    def path(self) -> Path:
        openings_dir = self.paths.openings_dir
        return (openings_dir / str(self)).with_suffix(".md")

    def exists(self) -> bool:
        return self.path.exists()

    @property
    def apy_header(self) -> str:
        return "model: Chessli Openings\ntags: chess::openings\ndeck: Chessli::openings\nmarkdown: False\n\n"

    def store(self, force: bool = False) -> None:
        if not self.exists() or force:
            log.info(f"Storing opening: {in_bold(str(self))}")
            md = f"{self.apy_header}"
            md += f"{self.md}\n\n"
            self.path.write_text(md)
        else:
            log.info(
                f"Ignoring {in_bold(str(self), 'yellow')}. You already know that opening :)"
            )

    def ankify(self):
        if self.exists():
            import_to_anki_via_apy(file_path=self.path)
        else:
            console.log(
                "To ankify, you first need to store the opening with `opening.store()`"
            )


@dataclass
class OpeningsCollection:
    config: Optional[DictConfig] = None
    paths: Optional[ChessliPaths] = None
    openings: Optional[List[Opening]] = None

    @classmethod
    def from_games(
        cls, config: DictConfig, paths: ChessliPaths, games: List["Game"]
    ) -> "OpeningsCollection":
        return cls(config, paths, [game.opening for game in games])

    def get_df(self):
        return pd.DataFrame(data=[opening.anki_items for opening in self.openings])

    def empty(self) -> bool:
        return not bool(len(self.openings))

    def export_csv(self) -> None:
        if self.empty():
            return
        time_stamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        openings_export_path = (
            self.paths.openings_dir / f"openings_export_{time_stamp}.csv"
        )
        self.get_df().to_csv(path_or_buf=openings_export_path, index=False)
        log.info(f"Exported openings as csv at {openings_export_path}")

        self.store_openings()

    def store_openings(self) -> None:
        if self.empty():
            return None
        for opening in self.openings:
            opening.store()
        log.info(f"Stored openings at {self.paths.openings_dir}")

    def ankify_openings(self) -> None:
        if self.empty():
            return None
        for opening in self.openings:
            opening.store()
            log.info(f"Ankifying opening: {in_bold(opening.name)}")
            opening.ankify()


def list_known_openings(
    eco_volume: Optional[ECOVolumeLetter], chessli_paths: ChessliPaths
) -> None:
    opening_dict = defaultdict(list)
    known_openings = sorted([f.stem for f in chessli_paths.openings_dir.glob("*.md")])
    print(
        f":fire: You already know a total of {in_bold(len(known_openings))} openings!!! :fire:",
        end="\n\n",
    )

    for opening in known_openings:
        opening_dict[opening[0]].append(opening)

    for key, value in opening_dict.items():
        if eco_volume is None or key == eco_volume:
            eco_volume_title = f"{ECOVolume[key]} ({len(value)})"
            print(f"[bold blue]{eco_volume_title}[/bold blue]!", end="\n\n")
            for val in value:
                print("✔️ ", val)
            print("")


def print_openings(openings: List["Opening"]):
    table = Table("", "Name", "ECO", title="New Openings")

    for opening in openings:
        if opening.exists():

            new_str = ""
            name_str = f"[grey]{opening.name}[/grey]"
            eco_str = f"[grey]{opening.eco}[/grey]"
        else:
            new_str = ":new:"
            name_str = f"[green]{opening.name}[/green]"
            eco_str = f"[green]{opening.eco}[/green]"

        table.add_row(new_str, eco_str, name_str)

    console.print(table)


@dataclass
class OpeningExtractorMixin:
    pgn: Optional[chess.pgn.Game] = None
    config: Optional[DictConfig] = None
    paths: Optional[ChessliPaths] = None

    @property
    def opening(self):
        game = self.pgn
        info = game.headers

        def get_moves(game) -> str:
            moves = []
            board = game.board()
            game = game.next()
            while info["Opening"] not in game.comment:
                move = game.move
                moves.append(move)
                game = game.next()
            move = game.move
            moves.append(move)
            move_list = board.variation_san(moves)
            return move_list

        return Opening(
            name=info["Opening"],
            eco=info["ECO"],
            site=info["Site"],
            moves=get_moves(game),
            config=self.config,
            paths=self.paths,
        )



import datetime
from typing import Optional

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import typer
from rich import print
from rich.console import Console

from chessli import users_client
from chessli.enums import PerfType
from chessli.utils import as_title, create_config_from_options

console = Console()
app = typer.Typer()


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context,):
    """Get stats and infos from Lichess"""

    ctx.params = ctx.parent.params
    print(f"{as_title('lichess stats')}", end="\n\n")


@app.command()
def leaderboard(
    ctx: typer.Context,
    type: Optional[PerfType] = typer.Option(
        None, help="Select which leaderboard you want to see"
    ),
):
    """Show (selected) lichess leaderboards"""

    config = create_config_from_options({**ctx.parent.params, **ctx.params})

    def print_lb(type, lb):
        print(f"[bold][blue]{type.upper()}[/bold][/blue]", end="\n\n")
        for rank, who in enumerate(lb):
            username = who["username"]
            title = who.get("title", "")
            rating = who["perfs"][type]["rating"]
            progress = who["perfs"][type]["progress"]
            if int(progress) >= 0:
                progress_str = f"([green]+{progress}[/green])"
            else:
                progress_str = f"([red]{progress}[/red])"
            print(f"{rank}. {username} ({title}) {rating} {progress_str}")

    if type is not None:
        print_lb(type, users_client.get_leaderboard(type))
    else:
        for key, value in users_client.get_all_top_10().items():
            print_lb(key, value)
            print("")


@app.command()
def rating(
    ctx: typer.Context,
    type: Optional[PerfType] = typer.Option(
        "blitz", help="Select type for which you want to see your rating"
    ),
):
    """Visualize the user's rating for selected chess game types"""
    config = create_config_from_options({**ctx.parent.params, **ctx.params})
    ratings = users_client.get_rating_history(config.user)
    available_types = [rating["name"].lower() for rating in ratings]
    if type not in available_types:
        print(
            f"User '{config.user}' did not play any '{type}' games yet! Time to play! :fire:"
        )

    for rating in ratings:
        name = rating["name"]
        if type is None or type == name.lower():
            ratings = [entry.rating for entry in rating["points"]]
            dates = [
                datetime.date(year=entry.year, month=entry.month + 1, day=entry.day)
                for entry in rating["points"]
            ]

            if ratings:
                console.log(f"Plotting {config.user}`s rating for '{type}' games")
                df = pd.DataFrame(index=dates, columns=[name], data=ratings)
                sns.set_theme(style="darkgrid")
                sns.lineplot(data=df, palette="tab10", linewidth=2.5)
                plt.show()
            else:
                print(
                    f"User '{config.user}' did not play any '{type}' games yet! Time to play! :fire:"
                )



from typing import List, Optional

import typer
from omegaconf import OmegaConf
from rich import print
from rich.console import Console
from rich.table import Table

from chessli.cli.option_callbacks import since_callback
from chessli.enums import PerfType, SinceEnum
from chessli.games import GamesFetcher, GamesReader
from chessli.openings import ECOVolumeLetter, OpeningsCollection, list_known_openings
from chessli.utils import (
    as_title,
    convert_since_enum_to_millis,
    create_config_from_options,
    extract_context_info,
    in_bold,
)

app = typer.Typer()
console = Console()


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context,):
    """Show and ankify chess openings"""

    ctx.params = ctx.parent.params
    print(as_title("chessli openings"), end="\n\n")


@app.command()
def ls(
    ctx: typer.Context,
    eco: Optional[ECOVolumeLetter] = typer.Option(
        default=None, help="Limit the shown openings to specific ECO volume"
    ),
    perf_type: Optional[List[PerfType]] = typer.Option(
        None, help="Filter fetching of games to the selected `perf_types`"
    ),
):
    """List your played openings"""
    chessli_paths, cli_config = extract_context_info(ctx)
    list_known_openings(eco, chessli_paths)


@app.command()
def ankify(
    ctx: typer.Context,
    new_openings_only: bool = typer.Option(True, help="Only ankify new openings"),
    since_enum: SinceEnum = typer.Option(
        SinceEnum.last_time,
        "--since",
        help="Filter fetching of games to those played since `since`",
        callback=since_callback,
    ),
    max: Optional[int] = typer.Option(30, help="Limit fetching of games to `max`",),
    perf_type: Optional[List[PerfType]] = typer.Option(
        None, help="Filter fetching of games to the selected `perf_types`"
    ),
    export_only: bool = typer.Option(
        True,
        "--export-only/--directly",
        help="Select to only export the created anki cards",
    ),
):
    """Parse your games to find new openings and create Anki cards"""
    chessli_paths, cli_config = extract_context_info(ctx)
    cli_config["since_millis"] = convert_since_enum_to_millis(
        since_enum, chessli_paths.user_config
    )

    if new_openings_only:
        games = GamesFetcher(chessli_paths, cli_config).fetch_games()
    else:
        games = GamesReader(chessli_paths, cli_config).games

    openings_collection = OpeningsCollection.from_games(
        config=cli_config, paths=chessli_paths, games=games
    )

    if export_only:
        openings_collection.export_csv()
    else:
        openings_collection.ankify_openings()


if __name__ == "__main__":
    app()



import collections
import io
import time
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import chess.pgn
import chess.svg
import pandas as pd
from omegaconf import DictConfig, OmegaConf
from rich import print
from rich.console import Console
from rich.progress import track
from rich.table import Table

from chessli import AnkifyError, ChessliPaths, berserk_client
from chessli.enums import Color
from chessli.mistakes import Mistake, MistakeFinderMixin, get_nag_name
from chessli.openings import Opening, OpeningExtractorMixin
from chessli.rich_logging import log
from chessli.utils import import_to_anki_via_apy, in_bold

console = Console()


games_client = berserk_client.games


@dataclass
class GamesFetcher:
    paths: ChessliPaths
    config: DictConfig

    def _print_games_table(self, counter: collections.Counter) -> None:
        games_table = self.counter_to_table(
            counter=counter,
            title="New Games :fire:",
            columns=["Game Type", "Number of Plays"],
        )
        console.log(games_table)

    def _print_openings_table(self, counter: collections.Counter) -> None:
        openings_table = self.counter_to_table(
            counter=counter,
            title="New Openings :fire:",
            columns=["Opening Name", "Number of Plays"],
        )
        console.log(openings_table)

    def _store_user_config(self) -> None:
        self.config.fetch_time = str(datetime.now())
        new_fetch_config = OmegaConf.create(
            {
                "since_millis": self.config.since_millis,
                "last_fetch_time": self.config.fetch_time,
            }
        )
        OmegaConf.save(config=new_fetch_config, f=self.paths.user_config_path)
        log.info(
            f"Updated the user config`s {in_bold('last_fetch_time')} with the current timestamp."
        )

    def _fetch_games_by_user(self) -> List:
        console.log(f"Fetching games of {in_bold(self.config.user)}.")
        games_by_user = []
        if self.config.perf_type is None:
            self.config.perf_type = [None]
        for perf_type in self.config.perf_type:
            if perf_type is not None:
                log.info(f"Fetching {in_bold(perf_type)} games...")
            for game in games_client.export_by_player(
                max=self.config.max,
                username=self.config.user,
                since=int(self.config.since_millis),
                perf_type=perf_type,
            ):
                games_by_user.append(game)
            time.sleep(1)

        return games_by_user

    def fetch_games(self):
        log.debug(self.config)

        new_games = []
        perftype_counter = collections.Counter()
        opening_counter = collections.Counter()

        games_by_user = self._fetch_games_by_user()

        if games_by_user:
            for game_json, game_pgn in [
                (
                    game,
                    games_client.export(
                        game["id"],
                        literate="true",
                        as_pgn=True,
                        opening="true",
                        tags="true",
                        moves="true",
                        evals="true",
                    ),
                )
                for game in games_by_user
            ]:
                game = self.pgn_to_game(game_pgn)
                chessli_game = Game(
                    pgn=game, json=game_json, config=self.config, paths=self.paths
                )

                if self.config.store:
                    log.debug(f"Storing {in_bold(chessli_game.name)}...")
                    chessli_game.store()

                new_games.append(chessli_game)

                perftype_counter.update([game_json["perf"]])
                opening_counter.update([game.headers["Opening"]])

            if self.config.verbosity >= 1:
                self._print_games_table(perftype_counter)
                self._print_openings_table(opening_counter)

            self._store_user_config()

        else:
            console.log(
                f"No new games to fetch since {in_bold(self.config.since_enum)}. Time to play! :fire:"
            )
            new_games = []
        return new_games

    @staticmethod
    def pgn_to_game(pgn: str) -> Optional[chess.pgn.Game]:
        return chess.pgn.read_game(io.StringIO(pgn))

    @staticmethod
    def counter_to_table(counter: collections.Counter, title: str, columns: List[str]):
        table = Table(*columns, title=f"{in_bold(title)}]")
        for game, count in counter.items():
            table.add_row(game, str(count))
        return table


@dataclass
class GamesReader:
    paths: ChessliPaths
    config: DictConfig

    def __post_init__(self):
        self._games = self.read_games()

    @property
    def path(self):
        return self.paths.games_dir

    @property
    def game_files(self) -> List[Path]:
        if self.config.perf_type is not None:
            game_path = self.path / self.config.perf_type
        else:
            game_path = self.path
        return list(game_path.glob("**/*.pgn"))

    @property
    def games(self) -> List[Any]:
        return self._games

    def read_games(self):
        games = [
            Game(
                pgn=chess.pgn.read_game(io.StringIO(pgn.read_text())),
                json=OmegaConf.load(pgn.with_suffix(".json")),
                config=self.config,
                paths=self.paths,
            )
            for pgn in self.game_files
        ]
        return games

    def ls(self) -> None:
        for game in self.games:
            print(game)

    @property
    def last_game(self) -> "Game":
        return self._games[-1]


@dataclass
class Game(MistakeFinderMixin, OpeningExtractorMixin, object):
    config: Optional[DictConfig] = None
    paths: Optional[ChessliPaths] = None
    pgn: Optional[chess.pgn.Game] = None
    json: Optional[Union[Dict, DictConfig]] = None

    def __str__(self) -> str:
        return f" - {self.name}"

    @property
    def name(self):
        return f"{self.pgn.headers['White']}_vs_{self.pgn.headers['Black']}_{self.json['id']}"

    @property
    def path(self) -> Path:
        path = self.paths.games_dir / self.json["perf"]
        path.mkdir(exist_ok=True)
        return path

    def store(self, as_pgn: bool = True, as_json: bool = True) -> None:
        if as_pgn:
            (self.path / self.name).with_suffix(".pgn").write_text(
                str(self.pgn), encoding="utf_8"
            )
        if as_json:
            for key, value in self.json.items():
                if isinstance(value, datetime):
                    self.json[key] = str(value)
            OmegaConf.save(
                config=OmegaConf.create(self.json),
                f=(self.path / self.name).with_suffix(".json"),
            )


@dataclass
class GamesCollection:
    games: List[Game]
    config: Optional[DictConfig]
    paths: Optional[ChessliPaths]
    mistakes: List[Mistake] = field(init=False)

    def __post_init__(self) -> None:
        self.mistakes = [mistakes for game in self.games for mistakes in game.mistakes]

    def get_df(self):
        return pd.DataFrame(data=[mistake.items for mistake in self.mistakes])

    def export_csv(self) -> None:
        time_stamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        games_export_path = (
            self.paths.mistakes_dir / f"games_mistakes_export_{time_stamp}.csv"
        )
        self.get_df().to_csv(path_or_buf=games_export_path, index=False)
        log.info(f"Exported games mistakes as `csv` at {in_bold(games_export_path)}")

    def ankify_games(self) -> None:
        for game in track(self.games, description="Ankifying your mistakes...\n"):
            num_mistakes = len(game.mistakes)
            log.info(
                f"Found {in_bold(num_mistakes, 'red')} mistakes in the game {in_bold(game.name)}."
            )
            if num_mistakes == 0:
                continue
            mistake_file_path = (self.paths.mistakes_dir / game.name).with_suffix(".md")

            apy_header = "model: Chessli Games\ntags: chess::game_analysis\ndeck: Chessli::games\nmarkdown: False\n\n"

            md_notes = [mistake.md for mistake in game.mistakes]
            md = f"{apy_header}"
            for md_note in md_notes:
                md += f"{md_note}\n\n"
            mistake_file_path.write_text(md)

            import_to_anki_via_apy(file_path=mistake_file_path)



import logging

from rich.logging import RichHandler

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True, markup=True)],
)

log = logging.getLogger("rich")



import sys
from enum import Enum
from typing import Optional

import typer
from rich import print
from rich.console import Console

from chessli import ChessliPaths, __version__, main_config
from chessli.cli import games as games_cli
from chessli.cli import openings as openings_cli
from chessli.cli import stats as stats_cli
from chessli.cli import tactics as tactics_cli
from chessli.cli.option_callbacks import version_callback
from chessli.rich_logging import log
from chessli.utils import as_title, in_bold

app = typer.Typer()
console = Console()


app.add_typer(openings_cli.app, name="openings")
app.add_typer(games_cli.app, name="games")
app.add_typer(stats_cli.app, name="stats")
app.add_typer(tactics_cli.app, name="tactics")


class LogLevel(Enum):
    debug = 10
    info = 20
    warning = 30


def log_level_from_verbosity(v: int) -> LogLevel:
    return {1: LogLevel.warning, 2: LogLevel.info, 3: LogLevel.debug}.get(
        v, LogLevel.info
    )


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    version: Optional[bool] = typer.Option(
        None, "--version", callback=version_callback, is_eager=True
    ),
    verbosity: int = typer.Option(
        2,
        "--verbose",
        "-v",
        count=True,
        help="Select verbosity level: Warning(-v), Info(-vv) Debug(-vvv), ",
    ),
    user: Optional[str] = typer.Option(None, help="Select a user name"),
    show_configs: bool = typer.Option(False, help="Show chessli configuration"),
    show_paths: bool = typer.Option(False, help="Show chessli paths"),
):
    f"""Chessli version {__version__}"""

    log_level = log_level_from_verbosity(verbosity).value
    log.setLevel(log_level)

    if main_config.token:
        log.debug(
            f"Chessli found your token. You'll be able to perform any request :fire:"
        )
    else:
        log.debug(
            f"Chessli did not found any lichess API token. You'll not be able to perform some requests."
        )

    if user is None:
        if main_config.user is None:
            console.log(
                f"You haven't chosen any username!\n Use `--user <your_username>`\n"
                f"Take a look at the documentation: https://pwenker.com/chessli/tutorial/ to learn how to set up a default username!"
            )
            sys.exit(1)
        else:
            user = main_config.user

    ctx.params["user"] = user
    ctx.params["paths"] = chessli_paths = ChessliPaths(user_name=user)

    if show_paths or log_level == LogLevel.debug:
        log.info(chessli_paths)

    if show_configs or log_level == LogLevel.debug:
        log.info(f"{in_bold('General Config')}")
        log.info(chessli_paths.main_config)
        log.info(f"{in_bold('User Config')}")
        log.info(chessli_paths.user_config)


if __name__ == "__main__":
    app()



"""
A free and open source chess improvement app that combines the power of Lichess and Anki.
"""
__version__ = "0.2.1"


from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, Dict, List, Optional, Set, Tuple, Union

import appdirs
import berserk
from omegaconf import DictConfig, OmegaConf
from rich.console import Console

from chessli.rich_logging import log

console = Console()

__all__ = ["ChessliPaths", "main_config", "berserk_client", "users_client"]


def get_berserk_client(token):
    if token is not None:
        session = berserk.TokenSession(token)
        return berserk.Client(session=session)
    else:
        return berserk.Client()


class AnkifyError(RuntimeError):
    """
    Could not ankify. Have you set up `apy` correctly?
    """


@dataclass
class PathCreatorMixin:
    @staticmethod
    def _maybe_make_dirs(dirs: List[Path]) -> None:
        for dir in dirs:
            dir.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def _maybe_touch_files(files: List[Path]) -> None:
        for file in files:
            file.touch(exist_ok=True)


@dataclass
class ChessliPaths(PathCreatorMixin, object):
    user_name: str
    data_dir: Path = Path(appdirs.user_data_dir("chessli"))
    configs_dir: Path = Path(appdirs.user_config_dir("chessli"))
    main_config_path: Path = field(init=False)
    user_data_dir: Path = field(init=False)
    user_configs_dir: Path = field(init=False)
    user_config_path: Path = field(init=False)
    games_dir: Path = field(init=False)
    openings_dir: Path = field(init=False)
    mistakes_dir: Path = field(init=False)
    tactics_dir: Path = field(init=False)

    def __post_init__(self) -> None:
        # Main paths
        self._maybe_make_dirs([self.data_dir, self.configs_dir])
        self.main_config_path = self.configs_dir / "config.yml"
        self._maybe_touch_files([self.main_config_path])

        # User-specific paths
        self.user_data_dir = self.data_dir / self.user_name
        self.user_configs_dir = self.configs_dir / self.user_name
        self.games_dir = self.user_data_dir / "games"
        self.openings_dir = self.user_data_dir / "openings"
        self.mistakes_dir = self.user_data_dir / "mistakes"
        self.tactics_dir = self.user_data_dir / "tactics"

        self._maybe_make_dirs(
            [
                self.user_data_dir,
                self.user_configs_dir,
                self.games_dir,
                self.openings_dir,
                self.mistakes_dir,
                self.tactics_dir,
            ]
        )
        self.user_config_path = self.user_configs_dir / "config.yml"
        self._maybe_touch_files([self.user_config_path])

        self._user_config = OmegaConf.load(self.user_config_path)
        self._main_config = OmegaConf.load(self.main_config_path)

    @property
    def main_config(self):
        return self._main_config

    @property
    def user_config(self):
        return self._user_config

    @user_config.setter
    def set_user_config(self, config):
        self._user_config = config

    @main_config.setter
    def set_main_config(self, config):
        self._main_config = config

    def __str__(self) -> str:
        return f"""
        Configs Directory: {self.configs_dir}
        Data Directory: {self.data_dir}
        General Config: {self.main_config_path}

        User Name: {self.user_name}
        User Configs Directory: {self.user_configs_dir}
        User Config Path: {self.user_config_path}
        User data directory: {self.user_data_dir}
        """


main_config = ChessliPaths("dummy").main_config
berserk_client = get_berserk_client(main_config.token)
users_client = berserk_client.users



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chessli', 'chessli.cli']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'berserk>=0.10.0,<0.11.0',
 'matplotlib>=3.3.3,<4.0.0',
 'omegaconf>=2.0.6,<3.0.0',
 'pandas>=1.2.0,<2.0.0',
 'python-chess>=1.999,<2.0',
 'rich>=9.9.0,<10.0.0',
 'seaborn>=0.11.1,<0.12.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['chessli = chessli.cli.main:app']}

setup_kwargs = {
    'name': 'chessli',
    'version': '0.2.1',
    'description': 'A free and open source chess improvement app that combines the power of lichess and anki',
    'long_description': "![Thumbnail](https://github.com/pwenker/chessli/blob/main/imgs/chessli.png?raw=true)\n\n_A free and open-source chess improvement app that combines the power of Lichess and Anki._\n\n|  | Chessli |\n| --- | --- |\n| Project                | ![GitHub Repo stars](https://img.shields.io/github/stars/pwenker/chessli?style=social) ![GitHub code size in bytes](https://img.shields.io/github/languages/code-size/pwenker/chessli) ![Lines of code](https://img.shields.io/tokei/lines/github/pwenker/chessli)\n| Activity & Issue Tracking | ![GitHub last commit](https://img.shields.io/github/last-commit/pwenker/chessli) ![GitHub issues](https://img.shields.io/github/issues-raw/pwenker/chessli) ![GitHub closed issues](https://img.shields.io/github/issues-closed-raw/pwenker/chessli)  |\n| PyPI                      | ![PyPI](https://img.shields.io/pypi/v/chessli)                                                                                                                                  ![PyPI - Python Version](https://img.shields.io/pypi/pyversions/chessli) ![PyPI - Downloads](https://img.shields.io/pypi/dm/chessli) |\n| Build & Health                  | ![GitHub Workflow Status](https://img.shields.io/github/workflow/status/pwenker/chessli/CI) ![Codecov](https://img.shields.io/codecov/c/github/pwenker/chessli) |\n| Docs | ![Documentation Status](https://img.shields.io/badge/Docs-live-green) ![](https://img.shields.io/badge/Tutorial-active-brightgreen) |\n\n## Demos\n\n### CLI Demo (watch whole video on [Youtube](https://www.youtube.com/embed/XbD71Kq7cx4))\n\n![CLI DEMO GIF](https://github.com/pwenker/chessli/blob/main/imgs/chessli_cli_demo.gif?raw=true)\n\n### Anki Cards Demo (watch whole video on [Youtube](https://www.youtube.com/embed/aj-FqJhPyyA))\n\n![CLI CARDS GIF](https://github.com/pwenker/chessli/blob/main/imgs/chessli_cards_demo.gif?raw=true)\n\n## Documentation\nCheck out the [documentation](https://pwenker.com/chessli):\n\n|  | Tutorial |\n| --- | --- |\nSetup chessli | [How to set chessli up](tutorial/how_to_set_up_chessli)\nGames & Mistakes | [How to learn from your games](tutorial/how_to_learn_from_your_games.md)\nOpenings | [How to build an opening repertoire](tutorial/how_to_create_an_opening_repertoire.md)\nTactics | [How to get better at tactics](tutorial/how_to_get_better_at_tactics.md)\nAnki Cards | [How to use chessli's anki cards](tutorial/how_to_use_chesslis_anki_cards.md)\n\n\n## Features\n\n- **Automatically fetch your games** and played tactics puzzles from [`lichess`](https://www.lichess.org) via the [`berserk`](https://github.com/rhgrant10/berserk) python client for the Lichess API.!\n- **Find your mistakes** by parsing your games and analysing them with [`python-chess`](https://github.com/niklasf/python-chess).\n- **Build a simple opening repertoire and list your known openings**\n- **Spaced repetition & Retrieval Practice**: Automatically (via [`apy`](https://github.com/lervag/apy)) or manually (via csv export) add your game mistakes, your openings and your tackled lichess puzzles into [`Anki`](https://apps.ankiweb.net/)\n- **Stats and visualizations**: Seamlessly show `lichess leaderboards` or plot your `rating history`.\n\n\n## Getting Started\n\n### Installation\n1. Install `pip`. See [here](https://pip.pypa.io/en/stable/installing/) for help.\n\n2.  Install chessli with `pip`:\n```console\npip install chessli\n```\nThat's it!\n\n## Basic Usage\nTo get help about `chessli`s commands, open your console and type:\n```console\nchessli --help\n```\nThe same works for subcommands, e.g., :\n```console\nchessli games --help\n```\nYou can find an overview of all availabe cli-commands [here](https://pwenker.com/chessli/cli/) in\nthe docs.\n\n### Youtube-Video: CLI Demo\n\n[![Chessli CLI Demo](https://img.youtube.com/vi/XbD71Kq7cx4/0.jpg)](https://www.youtube.com/embed/XbD71Kq7cx4)\n\nThere is also a short video showing the `chessli`s Anki cards in action:\n\n### Youtube-Video: Anki Cards Demo\n[![Chessli Anki Cards Demo](https://img.youtube.com/vi/aj-FqJhPyyA/0.jpg)](https://www.youtube.com/embed/aj-FqJhPyyA)\n\n\n### Tutorial\nNow as you are familiar with the basics, you might want to walk through the tutorial to get the most\nout of `chessli`!\nStart with [setting up chessli](tutorial/how_to_set_up_chessli.md).\n\n## Acknowledgments\n\n|  | Acknowledgements |\n| --- | --- |\n| [`Lichess`](https://lichess.org) | A free, no-ads, open source chess server that let's everyone play chess! Think about whether to [become a patron](https://lichess.org/patron)! :) |\n| [`Anki`](https://apps.ankiweb.net/) | A free and open-source flashcard program using spaced-repetition, a technique from cognitive science for fast and long-lasting memorization.  I couldn't imagine learning without it anymore. |\n| [`Anki Card Templates`](https://ankiweb.net/shared/info/1082754005) | The interactive chess functionality on Chessli's anki cards is based from [these fantastic cards](https://ankiweb.net/shared/info/1082754005).|\n| [`python-chess`](https://github.com/niklasf/python-chess) | Most of the heavy lifting, e.g. parsing games, finding mistakes, extracting openings, etc. is done with `python-chess`. |\n| [`berserk`](https://github.com/rhgrant10/berserk) |  The communication between `lichess` and `chessli` is performed via `berserk` |\n| [`typer`](https://github.com/tiangolo/typer) | The `chessli` `cli` is built with the great `typer` |\n| [`rich`](https://github.com/willmcgugan/rich) | The rich colors and fancy tables are made possible by `rich` |\n| [`apy`](https://github.com/lervag/apy/) | Importing cards directly into anki without csv-export can be done via `apy` |\n",
    'author': 'Pascal Wenker',
    'author_email': 'pwenker@posteo.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pwenker/chessli',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



import subprocess
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import Callable, Dict, List, Optional, Set, Tuple, Union

import berserk
import pandas as pd
import typer
from omegaconf import DictConfig, OmegaConf
from rich import print
from rich.console import Console
from rich.table import Table

from chessli import AnkifyError, ChessliPaths
from chessli.enums import SinceEnum

console = Console()


def create_config_from_options(context_params: Dict) -> DictConfig:
    def _strip_non_compatible_elements(context_params: Dict) -> Dict:
        return {
            k: v for k, v in context_params.items() if not isinstance(v, ChessliPaths)
        }

    config = OmegaConf.create(_strip_non_compatible_elements(context_params))
    return config


def extract_context_info(ctx: typer.Context) -> Tuple["ChessliPaths", "DictConfig"]:
    chessli_paths = ctx.parent.params["paths"]
    cli_params = {**ctx.parent.params, **ctx.params}
    try:
        cli_params["perf_type"] = (
            None if not cli_params["perf_type"] else cli_params["perf_type"]
        )
    except KeyError:
        pass
    cli_config = create_config_from_options(cli_params)
    return chessli_paths, cli_config


def convert_since_enum_to_millis(since_enum: SinceEnum, config: DictConfig):
    if since_enum == SinceEnum.one_hour:
        since = datetime.now() - timedelta(hours=1)
    elif since_enum == SinceEnum.yesterday:
        since = datetime.now() - timedelta(days=1)
    elif since_enum == SinceEnum.last_week:
        since = datetime.now() - timedelta(days=7)
    elif since_enum == SinceEnum.last_month:
        since = datetime.now() - timedelta(days=31)
    elif since_enum == SinceEnum.last_year:
        since = datetime.now() - timedelta(days=365)
    elif since_enum == SinceEnum.forever:
        since = datetime.now() - timedelta(days=365 * 42)  # roughly forever
    elif since_enum == SinceEnum.last_time:
        since = datetime.fromisoformat(config.last_fetch_time)
    return berserk.utils.to_millis(since)


def import_to_anki_via_apy(file_path: Path) -> None:
    try:
        subprocess.run(
            ["apy", "add-from-file", file_path], shell=True, check=True, input=b"n"
        )
    except subprocess.CalledProcessError as e:
        raise AnkifyError(e)


####################################################################################################
#                          Some custom functions to make strings prettier                          #
####################################################################################################


def in_bold(string: Union[str, int, Path], color: str = "blue") -> str:
    return f"[bold][{color}]{string}[/{color}][/bold]"


def as_title(string: str) -> str:
    return f":fire: {in_bold(string.upper())} :fire:"
